<?php

namespace Drupal\graphql_compose;

use Drupal\Core\Render\RenderContext;
use Symfony\Component\String\Inflector\EnglishInflector;
use function Symfony\Component\String\u;

class DataManager {

  protected $entityTypeManager = NULL;
  protected $entityFieldManager = NULL;
  protected $renderer = NULL;

  protected $inflector = NULL;
  protected $settings = [];
  protected $sdl = [];
  protected $definitions = [];

  public function __construct() {
    // Inject services from container.
    $this->entityTypeManager = \Drupal::service('entity_type.manager');
    $this->entityFieldManager = \Drupal::service('entity_field.manager');
    $this->renderer = \Drupal::service('renderer');
    // Inject services from container.

    $this->inflector = new EnglishInflector();

    // @TODO Define config and schemas and GUI to allow user select whict data to expose
    $this->settings = [
      'generate' => [
        'queryies' => true,
        'mutations' => false,
      ],
      'node' => [
        'interface' => ['Node'],
        'prefix' => 'Node',
        'storage_type' => 'node_type',
        'case' => 'camel', // 'none' | 'camel' | 'snake'
        'fields' => [
          // 'created',
          // 'changed',
          // 'info',          
          // 'promote',
          // 'sticky',
          // 'metatag',
          // 'path',
          // 'menu_link',
          'uuid' => [
            'type' => 'uuid',
            'label' => '{field_type}',
            'description' => 'The unique identifier for the {entity_type}.',
            'name_sdl' => 'id',
            'type_sdl' => 'ID',
            'isUnion' => FALSE,
            'isMultiple' => FALSE,
            'isRequired' => TRUE,
            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => 'entity_uuid',
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'entity',
                    'value' => NULL,
                  ],
                ],
              ],
            ]
          ],
          'uid' => [
            'type' => 'uid',
            'label' => '{field_type}',
            'description' => 'The author of the {entity_type}.',
            'name_sdl' => 'author',
            'type_sdl' => 'Actor',
            'isUnion' => FALSE,
            'isMultiple' => FALSE,
            'isRequired' => TRUE,
            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => 'entity_owner',
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'entity',
                    'value' => NULL,
                  ],
                ],
              ],
            ],
          ],
          'title' => [
            'type' => 'title',
            'label' => '{field_type}',
            'description' => 'The display title of the {entity_type}.',
            'name_sdl' => 'title',
            'type_sdl' => 'String',
            'isUnion' => FALSE,
            'isMultiple' => FALSE,
            'isRequired' => TRUE,
            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => 'entity_label',
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'entity',
                    'value' => NULL,
                  ],
                ],
              ],
            ]
          ]
        ],
      ],
      'user' => [
        'interface' => ['Node', 'Actor'],
        'prefix' => '', // 'User'
        'storage_type' => 'user',
        'case' => 'camel', // 'none' | 'camel' | 'snake'
        'defaults' => [
          'isUnion' => FALSE,
          'isMultiple' => FALSE,
          'isRequired' => FALSE,
        ],
        'fields' => [
          // uid
          // uuid
          // langcode
          // preferred_langcode
          // preferred_admin_langcode
          // name
          // pass
          // mail
          // timezone
          // status
          // created
          // changed
          // access
          // login
          // init
          // roles
          // default_langcode
          // metatag
          // path
          // user_picture
          'uuid' => [
            'type' => 'uuid',
            'label' => '{field_type}',
            'description' => 'The unique identifier for the {entity_type}.',
            'name_sdl' => 'id',
            'type_sdl' => 'ID',
            'isUnion' => FALSE,
            'isMultiple' => FALSE,
            'isRequired' => TRUE,
            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => 'entity_uuid',
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'entity',
                    'value' => NULL,
                  ],
                ],
              ],
            ]
          ],
          'name' => [
            'type' => 'String',
            'label' => 'The display name of the user.',
            'description' => 'The specific format of the display name could depend on permissions of the requesting user or application.',
            'name_sdl' => 'displayName',
            'type_sdl' => 'String',
            'isUnion' => FALSE,
            'isMultiple' => FALSE,
            'isRequired' => TRUE,
            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => 'entity_label',
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'entity',
                    'value' => NULL,
                  ],
                ],
              ],
            ]
          ],
          'mail' => [
            'type' => 'String',
            'label' => 'The e-mail of the user.',
            'description' => 'Can be null if the user has not filled in an e-mail or if the user/application making the request is not allowed to view this user\'s e-mail.',
            'name_sdl' => 'mail',
            'type_sdl' => 'Email',
            'isUnion' => FALSE,
            'isMultiple' => FALSE,
            'isRequired' => TRUE,
            'producers' => [
              [
                'type' => 'fromPath',
                'args' => [
                  'type' => 'entity:user',
                  'path' => '{field_name}.value',
                ],
              ],
            ],
          ],          
          'status' => [
            'type' => 'Integer',
            'label' => 'The status of the user account.',
            'description' => '',
            'name_sdl' => 'status',
            'type_sdl' => 'UserStatus',
            'isUnion' => FALSE,
            'isMultiple' => FALSE,
            'isRequired' => TRUE,
            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => 'user_status',
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'user',
                    'value' => NULL,
                  ],
                ],
              ],
            ]
          ],
          'roles' => [
            'type' => 'String',
            'label' => 'The roles that the user has.',
            'description' => '',
            'name_sdl' => 'roles',
            'type_sdl' => 'UserRole',
            'isUnion' => FALSE,
            'isMultiple' => TRUE,
            'isRequired' => TRUE,
            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => 'user_roles',
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'user',
                    'value' => NULL,
                  ],
                ],
              ],
            ]
          ],

        ]
      ],
      'media' => [
        'interface' => ['Node'],
        'prefix' => 'Media', // 'Media'
        'storage_type' => 'media_type',
        'case' => 'camel', // 'none' | 'camel' | 'snake'
        'defaults' => [
          'isUnion' => FALSE,
          'isMultiple' => FALSE,
          'isRequired' => FALSE,
        ],
        'fields' => [
          'uuid' => [
            'type' => 'uuid',
            'label' => '{field_type}',
            'description' => 'The unique identifier for the {entity_type}.',
            'name_sdl' => 'id',
            'type_sdl' => 'ID',
            'isUnion' => FALSE,
            'isMultiple' => FALSE,
            'isRequired' => TRUE,
            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => 'entity_uuid',
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'entity',
                    'value' => NULL,
                  ],
                ],
              ],
            ]
          ],
        ]
      ],
      // @TODO validate custom module paragraphs is installed
      'paragraph'  => [
        'interface' => ['Node'],
        'prefix' => 'Paragraph',
        'storage_type' => 'paragraphs_type',
        'case' => 'camel', // 'none' | 'camel' | 'snake'
        'fields' => [
          // id
          // uuid
          // revision_id
          // langcode
          // type
          // status
          // created
          // parent_id
          // parent_type
          // parent_field_name
          // behavior_settings
          // default_langcode
          // revision_default
          // revision_translation_affected
          'uuid' => [
            'type' => 'uuid',
            'label' => '{field_type}',
            'description' => 'The unique identifier for the {entity_type}.',
            'name_sdl' => 'id',
            'type_sdl' => 'ID',
            'isUnion' => FALSE,
            'isMultiple' => FALSE,
            'isRequired' => TRUE,

            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => 'entity_uuid',
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'entity',
                    'value' => NULL,
                  ],
                ],
              ],
            ]
            
          ],
        ],
      ],
      // Core fields
      'fields' => [
        'prefix' => '', // 'field_'
        'case' => 'camel', // 'none' | 'camel' | 'snake'
        'defaults' => [
          'isUnion' => FALSE,
          'isMultiple' => FALSE,
          'isRequired' => FALSE,
        ],
        'types' => [

          'string' => [
            'type_sdl' => 'String',
            'producers' => [
              [
                'type' => 'fromPath',
                'args' => [
                  'type' => 'entity:node',
                  'path' => '{field_name}.value',
                ],
              ],
            ],
          ],

          'string_long' => [
            'type_sdl' => 'String',
            'producers' => [
              [
                'type' => 'fromPath',
                'args' => [
                  'type' => 'entity:node',
                  'path' => '{field_name}.value',
                ],
              ],
            ],
          ],

          'link' => [
            'type_sdl' => 'Link',
            'producers' => [

              [
                'type' => 'dataProducer',
                'id' => 'field_link',
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'entity',
                    'value' => NULL,
                  ],
                  [
                    'id' => 'fromValue',
                    'key' => 'field',
                    'value' => '{field_name}',
                  ]
                ],
              ],

            ],
          ],

          'media_image' => [
            'type_sdl' => 'MediaImage',
            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => '', // 'entity_reference' or `field_first` (depending on isMultiple aka cardinality)
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'entity',
                    'value' => NULL,
                  ],
                  [
                    'id' => 'fromValue',
                    'key' => 'field',
                    'value' => '{field_name}',
                  ],
                ] 
              ]
            ]
          ],

          'image' => [
            'type_sdl' => 'Image',
            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => 'property_path',
                'map' => [
                  [
                    'id' => 'fromValue',
                    'key' => 'type',
                    'value' => 'entity:file',
                  ],
                  [
                    'id' => 'fromParent',
                    'key' => 'value',
                    'value' => NULL,
                  ],
                  [
                    'id' => 'fromValue',
                    'key' => 'path',
                    'value' => '{field_name}.entity',
                  ],
                ]
              ],
              [
                'type' => 'dataProducer',
                'id' => 'image_derivative',
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'entity',
                    'value' => NULL,
                  ],
                  [
                    'id' => 'fromValue',
                    'key' => 'style',
                    'value' => 'wide', // @TODO privide GUI to enter value or set to default to no image_style
                  ]
                ],
              ],

            ],
          ],

          // @TODO provide date formatters
          'datetime' => [
            'type_sdl' => 'String',
            'producers' => [
              [
                'type' => 'fromPath',
                'args' => [
                  'type' => 'entity:node',
                  'path' => '{field_name}.value',
                ],
              ],
            ],
          ],

          // @TODO validate strategy
          'entity_reference' => [
            'type_sdl' => '', // this should be defined by target bundle(s) as Bundle Type or Union
            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => '', // 'entity_reference' or `field_first` (depending on isMultiple)
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'entity',
                    'value' => NULL,
                  ],
                  [
                    'id' => 'fromValue',
                    'key' => 'field',
                    'value' => '{field_name}',
                  ]
                ],
              ],
            ],
          ],

          'entity_reference_revisions' => [
            'type_sdl' => '', // this should be defined by target bundle(s) as Bundle Type or Union
            'producers' => [
              [
                'type' => 'dataProducer',
                'id' => '', // 'entity_reference' or `field_first` (depending on isMultiple)
                'map' => [
                  [
                    'id' => 'fromParent',
                    'key' => 'entity',
                    'value' => NULL,
                  ],
                  [
                    'id' => 'fromValue',
                    'key' => 'field',
                    'value' => '{field_name}',
                  ]
                ],
              ],
            ],
          ],
          
        ]
      ],
    ];

    $this->calculateEntity('user');
    $this->calculateSdl('user');

    $this->calculateEntity('media');
    $this->calculateSdl('media');

    $this->calculateEntity('paragraph');
    $this->calculateSdl('paragraph');

    $this->calculateEntity('node');
    $this->calculateSdl('node');
  }

  public function getDefinitions($type) {
    if (!array_key_exists($type, $this->definitions)) {
      return NULL;
    }

    return $this->definitions[$type];
  }

  public function getSdlByStorage($storage, $type) {
    if (!array_key_exists($storage, $this->sdl)) {
      return NULL;
    }

    if (!array_key_exists($type, $this->sdl[$storage])) {
      return NULL;
    }

    return $this->sdl[$storage][$type];
  }

  protected function getEntitiesFromStorageType($type) {
    $entities = [];
    if ($type === 'user') {
      $entities[] = [
        'id' => 'user',
        'label' => 'Users',
        'description' => 'The GraphQL API users.',
      ];

      return $entities;  
    }

    $storage = $this->settings[$type]['storage_type'];
    foreach ( $this->entityTypeManager->getStorage($storage)->loadMultiple() as $entity) {
      $entities[] = [
        'id' => $entity->id(),
        'label' => $entity->label(),
        'description' => $entity->getDescription(),
      ];
    }

    return $entities;
  }

  protected function calculateEntity($type) {
    $entities = $this->getEntitiesFromStorageType($type);
    $this->calculateFields($type, $entities);
  }

  protected function calculateSdl($type) {
    $context = new RenderContext();
    $base = '';
    $extension = '';
    foreach ($this->definitions[$type] as $entity) {
      $renderableBase = [
        '#theme' => 'entity_base',
        '#entity' => $entity,
      ];

      $base .= $this->renderer->executeInRenderContext($context, function() use (&$renderableBase) {
        return $this->renderer->render($renderableBase);
      });

      $renderableExtension = [
        '#theme' => 'entity_extension',
        '#entity' => $entity,
      ];

      $extension .= $this->renderer->executeInRenderContext($context, function() use (&$renderableExtension) {
        return $this->renderer->render($renderableExtension);
      });
    }

    $this->sdl[$type] = [
      'base' => $base,
      'extension' => $extension,
    ];
  }

  protected function calculateFields($entityType, $entities) {
    $fieldNames = $this->settings[$entityType]['fields']?:[];
    $fieldTypes = $this->settings['fields']['types']?:[];
    $types = [];

    foreach ($entities as [ 'id' => $entityId, 'label' => $entityLabel, 'description' => $entityDescription ] ) {
      $prefix = $this->settings[$entityType]['prefix'];
      $type = u($entityId)->title()->prepend($prefix)->camel()->toString();
      $fields = [];
      $unions = [];

      foreach ($this->entityFieldManager->getFieldDefinitions($entityType, $entityId) as $field) {
        if ($field instanceof BaseFieldDefinition) {
          continue;
        }

        if (!in_array($field->getName(), array_keys($fieldNames)) && !u($field->getName())->startsWith('field_')) {
          continue;
        }

        $fieldInfo = [];
        // Default fields
        if (in_array($field->getName(), array_keys($fieldNames))) {
          $entitySDL = u($entityLabel)->camel()->title()->toString();
          $fieldInfo =  $fieldNames[$field->getName()];
          $fieldInfo['label'] = u($fieldInfo['label'])->replace('{field_type}', $field->getName())->toString();
          $fieldInfo['name'] = $field->getName();
          $fieldInfo['description'] = u($fieldInfo['description'])->replace('{entity_type}', $entitySDL)->toString();
        }

        // Custom fields
        if (u($field->getName())->startsWith('field_')) {
          if (!in_array($field->getType(), array_keys($fieldTypes))) {
            continue;
          }

          $fieldName = u($field->getName())->trimPrefix('field_')->camel()->title()->toString();
          $fieldInfo = array_merge($fieldTypes[$field->getType()], $this->settings['fields']['defaults']);

          if ($field->getType() === 'entity_reference' || $field->getType() === 'entity_reference_revisions') {

            $handlerSettings = $field->getSetting('handler_settings');
            // @TODO fix ReusableParagraph that contains target_bundles with NULL value
            if ($handlerSettings || $handlerSettings['target_bundles'] === NULL) {
              continue;
            }

            $referenceTargetBundles = array_keys($handlerSettings['target_bundles']);
            $targetType = $field->getSetting('target_type');

            if (!$referenceTargetBundles) {
              continue;
            }

            if ( !array_intersect($referenceTargetBundles, array_keys($fieldTypes)?:[]) &&
                 !array_key_exists($targetType, $this->definitions) &&
                 !array_intersect($referenceTargetBundles, array_keys($this->definitions[$targetType]))
               ) {
                continue;
            }

            if (count($referenceTargetBundles) > 1) {
              $fieldInfo['isUnion'] = TRUE;
              $unionType = u($fieldName)->prepend($type)->title()->append('Union')->toString();
              $fieldInfo['type_sdl'] = $unionType;

              $unions[$unionType]['type'] = $targetType;
              $unions[$unionType]['type_sdl'] = $unionType;
              $unions[$unionType]['target_bundles'] = $referenceTargetBundles;

              foreach ($referenceTargetBundles as $referenceTargetBundle) {

                if (array_key_exists($referenceTargetBundle, $fieldTypes) &&  array_key_exists('type_sdl', $fieldTypes[$referenceTargetBundle])) {
                  $unions[$unionType]['mapping'][$referenceTargetBundle] = $fieldTypes[$referenceTargetBundle]['type_sdl'];
                }

                if ($this->definitions[$targetType][$referenceTargetBundle]['type_sdl']) {
                  $unions[$unionType]['mapping'][$referenceTargetBundle] = $this->definitions[$targetType][$referenceTargetBundle]['type_sdl'];
                }
              }
              $fieldInfo['producers'][0]['id'] = 'entity_reference';
            } else {
              $referenceTargetBundle = $targetType . '_' . reset($referenceTargetBundles);
              if (array_key_exists($referenceTargetBundle, $fieldTypes)) {
                $fieldInfo = array_merge($fieldTypes[$referenceTargetBundle], $this->settings['fields']['defaults']);
              }

              $referenceTargetBundle = reset($referenceTargetBundles);
              if ( $this->definitions[$targetType][$referenceTargetBundle] ) {
                $fieldInfo['type_sdl'] = $this->definitions[$targetType][$referenceTargetBundle]['type_sdl'];
              }
              
              $fieldInfo['producers'][0]['id'] = $field->getFieldStorageDefinition()->isMultiple() ? 'entity_reference' : 'field_first';
            }
          }

          $fieldInfo['type'] = $field->getType();
          $fieldInfo['name'] = $field->getName();
          $fieldInfo['label'] = $field->getLabel();
          $fieldInfo['description'] = $field->getDescription();
          $fieldInfo['name_sdl'] = u($field->getName())->trimPrefix('field_')->camel()->toString();
          $fieldInfo['isMultiple'] = $field->getFieldStorageDefinition()->isMultiple() || $fieldInfo['isUnion'];

          if (!$fieldInfo['producers']) {
            continue;
          }          
        }

        $fields[$field->getName()] = $fieldInfo;
      }

      $this->definitions[$entityType][$entityId] = [
        'id' => $entityId,
        'type' => $type,
        'interface' => $this->settings[$entityType]['interface']?:[],
        'label' => $entityLabel,
        'description' => $entityDescription,
        'type_plural' => $this->inflector->pluralize($type)[0],
        'type_sdl' => u($entityId)->camel()->title()->prepend($prefix)->toString(),
        'fields' => $fields,
        'unions' => $unions,
      ];
    }

    return;
  }
}